import typing as _typing

import aiofiles as _aiofiles
from aiofiles.threadpool.text import AsyncTextIOWrapper as _AsyncTextIOWrapper
from aiocsv import AsyncDictWriter as _AsyncDictWriter


async def create_csv(
    path: str,
    headers: _typing.Sequence[str]
) -> None:
    """Replace or create file at path with empty CSV with column headers.
    
    Parameters
    ----------
    path : str
        File path to CSV file.
    headers : Sequence[str]
        Column header names to be written to new CSV file.

    Returns
    -------
    None
    
    Raises
    ------
    FileNotFoundError
        If file path directory does not exist.
        
    Example
    -------
    >>> import asyncio
    >>>
    >>> from aiocsv_utils.write import create_csv
    >>>
    >>> asyncio.run(create_csv('data/people.csv', ['id', 'name', 'age', 'ssn']))
    >>>
    >>> with open('cities.csv', mode='r', encoding='utf-8', newline='') as f:
    >>>     print(f.read())
    id,name,age,ssn
    """
    async with _aiofiles.open(path, 'w') as f:
        dictwriter = _AsyncDictWriter(f, fieldnames=headers)
        await dictwriter.writeheader()


async def write_csv_file_row(
    async_file: _AsyncTextIOWrapper,
    record: dict[str, _typing.Any],
    headers: _typing.Sequence[str]
) -> None:
    """Append a record to a CSV file.
    
    Parameters
    ----------
    async_file : AsyncTextIOWrapper
        Async file object from aiofiles.read
    record : dict[str, Any]
        Record to be written to CSV file.
    headers : Sequence[str]
        CSV Column header names.

    Returns
    -------
    None
        
    Example
    -------
    >>> import asyncio
    >>> import aiofiles
    >>>
    >>> from aiocsv_utils.write import write_csv_file_row
    >>>
    >>> record = {'id': 1, 'name': 'John', 'age': 30, 'ssn': '111-22-3333'}
    >>> headers = ['id', 'name', 'age', 'ssn']
    >>> 
    >>> async def write_line(path, record, headers) -> None:
    >>>     async with aiofiles.open(path, 'w') as f:
    >>>         write_csv_file_row(f, record, headers)
    >>>
    >>> asyncio.run(write_line('data/people.csv', record, headers))
    >>>
    >>> with open('cities.csv', mode='r', encoding='utf-8', newline='') as f:
    >>>     print(f.read())
    id,name,age,ssn
    1,John,30,111-22-3333
    """
    dictwriter = _AsyncDictWriter(async_file, fieldnames=headers)
    await dictwriter.writerow(record)
        

async def write_csv_row(
    path: str,
    record: dict[str, _typing.Any],
    headers: _typing.Sequence[str]
) -> None:
    """Append a record to a CSV file.
    
    Parameters
    ----------
    path : str
        File path to CSV file.
    record : dict[str, Any]
        Record to be written to CSV file.
    headers : Sequence[str]
        CSV Column header names.

    Returns
    -------
    None
    
    Raises
    ------
    FileNotFoundError
        If file path does not exist.
        
    Example
    -------
    >>> import asyncio
    >>>
    >>> from aiocsv_utils.write import write_csv_row
    >>>
    >>> record = {'id': 1, 'name': 'John', 'age': 30, 'ssn': '111-22-3333'}
    >>> headers = ['id', 'name', 'age', 'ssn']
    >>> asyncio.run(write_csv_row('data/people.csv', record, headers))
    >>>
    >>> with open('cities.csv', mode='r', encoding='utf-8', newline='') as f:
    >>>     print(f.read())
    id,name,age,ssn
    1,John,30,111-22-3333
    """
    async with _aiofiles.open(path, 'a') as f:
        await write_csv_file_row(f, record, headers)
 

async def write_csv_file_rows(
    async_file: _AsyncTextIOWrapper,
    records: list[dict[str, _typing.Any]],
    headers: _typing.Sequence[str]
) -> None:
    """Append a list of records to a CSV file.
    
    Parameters
    ----------
    async_file : AsyncTextIOWrapper
        Async file object from aiofiles.read
    records : list[dict[str, Any]]
        List of records to be written to CSV file.
    headers : Sequence[str]
        CSV Column header names.

    Returns
    -------
    None
        
    Example
    -------
    >>> import asyncio
    >>>
    >>> from aiocsv_utils.write import write_csv_file_rows
    >>>
    >>> records = [{'id': 2, 'name': 'Jane', 'age': 25, 'ssn': '222-33-4444'},
    >>>            {'id': 3, 'name': 'Mike', 'age': 47, 'ssn': '000-11-2222'}]
    >>> headers = ['id', 'name', 'age', 'ssn']
    >>>
    >>> async def write_rows(path: str, records: list[dict], headers: list[str]) -> None:
    >>>     await write_csv_file_rows(path, records, headers)
    >>>
    >>> asyncio.run(write_rows('data/people.csv', records, headers))
    >>>
    >>> with open('cities.csv', mode='r', encoding='utf-8', newline='') as f:
    >>>     print(f.read())
    id,name,age,ssn
    1,John,30,111-22-3333
    2,Jane,25,222-33-4444
    3,Mike,47,000-11-2222
    """
    dictwriter = _AsyncDictWriter(async_file, fieldnames=headers)
    await dictwriter.writerows(records)
        
        
async def write_csv_rows(
    path: str,
    records: list[dict[str, _typing.Any]],
    headers: _typing.Sequence[str]
) -> None:
    """Append a list of records to a CSV file.
    
    Parameters
    ----------
    path : str
        File path to CSV file.
    records : list[dict[str, Any]]
        List of records to be written to CSV file.
    headers : Sequence[str]
        CSV Column header names.

    Returns
    -------
    None
    
    Raises
    ------
    FileNotFoundError
        If file path does not exist.
        
    Example
    -------
    >>> import asyncio
    >>>
    >>> from aiocsv_utils.write import write_csv_row
    >>>
    >>> records = [{'id': 2, 'name': 'Jane', 'age': 25, 'ssn': '222-33-4444'},
    >>>            {'id': 3, 'name': 'Mike', 'age': 47, 'ssn': '000-11-2222'}]
    >>> headers = ['id', 'name', 'age', 'ssn']
    >>> asyncio.run(write_csv_row('data/people.csv', records, headers))
    >>>
    >>> with open('cities.csv', mode='r', encoding='utf-8', newline='') as f:
    >>>     print(f.read())
    id,name,age,ssn
    1,John,30,111-22-3333
    2,Jane,25,222-33-4444
    3,Mike,47,000-11-2222
    """
    async with _aiofiles.open(path, 'a') as f:
        await write_csv_file_rows(f, records, headers)


import pytest

from aiocsv_utils.read import csv_headers
from aiocsv_utils.read import csv_to_records
from aiocsv_utils.read import csv_to_records_chunks


@pytest.mark.asyncio
async def test_csv_headers():
    expected = ['LatD', 'LatM', 'LatS', 'NS', 'LonD',
                'LonM', 'LonS', 'EW', 'City', 'State']
    assert await csv_headers('data/cities.csv') == expected
    
    
@pytest.mark.asyncio
async def test_empty_csv_headers():
    expected = ['LatD', 'LatM', 'LatS', 'NS', 'LonD',
                'LonM', 'LonS', 'EW', 'City', 'State']
    assert await csv_headers('data/cities_empty.csv') == expected
    

@pytest.mark.asyncio
async def test_csv_to_records():
    async for row in csv_to_records('data/cities.csv'):
        assert row == {'LatD': 41, 'LatM': 5, 'LatS': 59, 'NS': 'N', 'LonD': 80, 'LonM': 39,
                       'LonS': 0, 'EW': 'W', 'City': 'Youngstown', 'State': 'OH'}
        break
    
    
@pytest.mark.asyncio
async def test_empty_csv_to_records():
    with pytest.raises(StopAsyncIteration) as e_info:
        await anext(csv_to_records('data/cities_empty.csv'))
        
        
@pytest.mark.asyncio
async def test_csv_to_records_chunks():
    async for chunk in csv_to_records_chunks('data/cities.csv', 2):
        assert chunk == [
            {'LatD': 41, 'LatM': 5, 'LatS': 59, 'NS': 'N', 'LonD': 80, 'LonM': 39,
             'LonS': 0, 'EW': 'W', 'City': 'Youngstown', 'State': 'OH'},
            {'LatD': 42, 'LatM': 52, 'LatS': 48, 'NS': 'N', 'LonD': 97, 'LonM': 23,
             'LonS': 23, 'EW': 'W', 'City': 'Yankton', 'State': 'SD'}
        ]
        break


@pytest.mark.asyncio
async def test_empty_csv_to_records_chunks():
    with pytest.raises(StopAsyncIteration) as e_info:
        await anext(csv_to_records_chunks('data/cities_empty.csv', 2))


import typing as _typing

import aiofiles as _aiofiles
from aiofiles.threadpool.text import AsyncTextIOWrapper as _AsyncTextIOWrapper
import aiocsv as _aiocsv
from aioitertools.more_itertools import chunked as _chunked

from .convert import convert_str as _convert_str


async def csv_file_headers(
    async_file: _AsyncTextIOWrapper,
    delimiter=','
) -> list[str]:
    """Asynchronously read the header names of a csv file.
    
    Parameters
    ----------
    async_file : AsyncTextIOWrapper
        Async file object from aiofiles.read
    delimiter : str, optional
        CSV file delimiter character.

    Returns
    -------
    list[str]
        A list of the CSV column headers.
        
    Example
    -------
    >>> import asyncio
    >>> import aiofiles
    >>>
    >>> from aiocsv_utils.read import csv_file_headers
    >>>
    >>> async def get_headers(path: str) -> list[str]:
    >>>     async with aiofiles.open(path, mode='r', encoding='utf-8', newline='') as f:
    >>>         return csv_file_headers(f)
    >>>
    >>> asyncio.run(get_headers('data/cities.csv'))
    ['LatD', 'LatM', 'LatS', 'NS', 'LonD', 'LonM', 'LonS', 'EW', 'City', 'State']
    """
    reader = _aiocsv.AsyncReader(async_file, delimiter=delimiter)
    return await anext(reader)
        

async def csv_headers(
    path: str,
    mode='r',
    encoding='utf-8',
    newline='',
    delimiter=','
) -> list[str]:
    """Asynchronously read the header names of a csv by path.
    
    Parameters
    ----------
    path : str
        File path to CSV file.
    mode : str, optional
        Mode while opening a file.
    encoding : str, optional
        The encoding format.
    newline : str, optional
        How newlines mode works.
    delimiter : str, optional
        CSV file delimiter character.

    Returns
    -------
    list[str]
        A list of the CSV column headers.
        
    Raises
    ------
    FileNotFoundError
        If file does not exist.
        
    Example
    -------
    >>> import asyncio
    >>>
    >>> from aiocsv_utils.read import csv_headers
    >>>
    >>> asyncio.run(csv_headers('data/cities.csv'))
    ['LatD', 'LatM', 'LatS', 'NS', 'LonD', 'LonM', 'LonS', 'EW', 'City', 'State']
    """
    async with _aiofiles.open(path, mode=mode, buffering=-1, encoding=encoding, errors=None, newline=newline) as afp:
        return await csv_file_headers(afp, delimiter)


async def csv_file_to_records(
    async_file: _AsyncTextIOWrapper,
    delimiter=','
) -> _typing.AsyncGenerator[dict[str, _typing.Any], None]:
    """Asynchronously read a CSV file and async yield each record.
    
    Parameters
    ----------
    async_file : AsyncTextIOWrapper
        Async file object from aiofiles.read
    delimiter : str, optional
        CSV file delimiter character.

    Returns
    -------
    AsyncGenerator[dict[str, Any]
        An async generator that yields each CSV row as a dict.
        
    Example
    -------
    >>> import asyncio
    >>> import aiofiles
    >>>
    >>> from aiocsv_utils.read import csv_file_to_records
    >>>
    >>> async def read_first_record() -> list[dict] | None:
    >>>     async with aiofiles.open(path, mode='r', encoding='utf-8', newline='') as f:
    >>>         async for row in csv_file_to_records(f):
    >>>             return row
    >>>
    >>> asyncio.run(read_first_record())
    {'LatD': 41, 'LatM': 5, 'LatS': 59, 'NS': 'N', 'LonD': 80, 'LonM': 39,
     'LonS': 0, 'EW': 'W', 'City': 'Youngstown', 'State': 'OH'}
    """
    async for row in _aiocsv.AsyncDictReader(async_file, delimiter=delimiter):
        yield {col: _convert_str(val) for col, val in row.items()}
            

async def csv_to_records(
    path: str,
    mode='r',
    encoding='utf-8',
    newline='',
    delimiter=','
) -> _typing.AsyncGenerator[dict[str, _typing.Any], None]:
    """Asynchronously read a CSV by path and async yield each record.
    
    Parameters
    ----------
    path : str
        File path to CSV file.
    mode : str, optional
        Mode while opening a file.
    encoding : str, optional
        The encoding format.
    newline : str, optional
        How newlines mode works.
    delimiter : str, optional
        CSV file delimiter character.

    Returns
    -------
    AsyncGenerator[dict[str, Any]
        An async generator that yields each CSV row as a dict.
        
    Raises
    ------
    FileNotFoundError
        If file does not exist.
        
    Example
    -------
    >>> import asyncio
    >>>
    >>> from aiocsv_utils.read import csv_to_records
    >>>
    >>> async def read_first_record() -> list[dict] | None:
    >>>     async for row in csv_to_records('data/cities.csv'):
    >>>         return row
    >>>
    >>> asyncio.run(read_first_record())
    {'LatD': 41, 'LatM': 5, 'LatS': 59, 'NS': 'N', 'LonD': 80, 'LonM': 39,
     'LonS': 0, 'EW': 'W', 'City': 'Youngstown', 'State': 'OH'}
    """
    async with _aiofiles.open(path, mode=mode, encoding=encoding, newline=newline) as afp:
        async for row in csv_file_to_records(afp, delimiter):
            yield row
            

async def csv_file_to_records_chunks(
    async_file: _AsyncTextIOWrapper,
    chunk_size: int,
    delimiter=','
) -> _typing.AsyncGenerator[list[dict[str, _typing.Any]], None]:
    """Asynchronously read a CSV file and async yield chunks of records.
    
    Parameters
    ----------
    async_file : AsyncTextIOWrapper
        Async file object from aiofiles.read
    delimiter : str, optional
        CSV file delimiter character.

    Returns
    -------
    AsyncGenerator[list[dict[str, Any]]]
        An async generator that yields chunks of CSV rows as lists of dicts.
        
    Example
    -------
    >>> import asyncio
    >>>
    >>> from aiocsv_utils.read import csv_file_to_records_chunks
    >>>
    >>> async def read_first_two_record() -> dict | None:
    >>>     async with aiofiles.open(path, mode='r', encoding='utf-8', newline='') as f:
    >>>         async for chunk in csv_file_to_records_chunks(f, 2):
    >>>             return chunk
    >>>
    >>> asyncio.run(read_first_two_record())
    [{'LatD': 41, 'LatM': 5, 'LatS': 59, 'NS': 'N', 'LonD': 80, 'LonM': 39,
    'LonS': 0, 'EW': 'W', 'City': 'Youngstown', 'State': 'OH'},
    {'LatD': 42, 'LatM': 52, 'LatS': 48, 'NS': 'N', 'LonD': 97, 'LonM': 23,
    'LonS': 23, 'EW': 'W', 'City': 'Yankton', 'State': 'SD'}]
    """
    records = csv_file_to_records(async_file, delimiter)
    async for chunk in _chunked(records, chunk_size):
        yield chunk
        
async def csv_to_records_chunks(
    path: str,
    chunk_size: int,
    mode='r',
    encoding='utf-8',
    newline='',
    delimiter=','
) -> _typing.AsyncGenerator[list[dict[str, _typing.Any]], None]:
    """Asynchronously read a CSV by path and async yield chunks of records.
    
    Parameters
    ----------
    path : str
        File path to CSV file.
    mode : str, optional
        Mode while opening a file.
    encoding : str, optional
        The encoding format.
    newline : str, optional
        How newlines mode works.
    delimiter : str, optional
        CSV file delimiter character.

    Returns
    -------
    AsyncGenerator[list[dict[str, Any]]]
        An async generator that yields chunks of CSV rows as lists of dicts.
    
    Raises
    ------
    FileNotFoundError
        If file does not exist.
        
    Example
    -------
    >>> import asyncio
    >>>
    >>> from aiocsv_utils.read import csv_to_records_chunks
    >>>
    >>> async def read_first_two_record() -> dict | None:
    >>>     async for chunk in csv_to_records_chunks('data/cities.csv', 2):
    >>>         return chunk
    >>>
    >>> asyncio.run(read_first_two_record())
    [{'LatD': 41, 'LatM': 5, 'LatS': 59, 'NS': 'N', 'LonD': 80, 'LonM': 39,
    'LonS': 0, 'EW': 'W', 'City': 'Youngstown', 'State': 'OH'},
    {'LatD': 42, 'LatM': 52, 'LatS': 48, 'NS': 'N', 'LonD': 97, 'LonM': 23,
    'LonS': 23, 'EW': 'W', 'City': 'Yankton', 'State': 'SD'}]
    """
    async with _aiofiles.open(path, mode=mode, encoding=encoding, newline=newline) as f:
        async for chunk in csv_file_to_records_chunks(f, chunk_size, delimiter):
            yield chunk


from aiocsv_utils.convert import convert_str


def test_convert_int():
    assert convert_str(1) == 1
    
def test_convert_str_int():
    assert convert_str('11') == 11
    
def test_convert_float():
    assert convert_str(1.1) == 1.1
    
def test_convert_str_float():
    assert convert_str('1.1') == 1.1
    
def test_convert_str():
    assert convert_str('NaN') == 'NaN'
    
def test_convert_true():
    assert convert_str(True) == True
    
def test_convert_str_true():
    assert convert_str('True') == True
    
def test_convert_false():
    assert convert_str(False) == False
    
def test_convert_str_false():
    assert convert_str('False') == False
    
def test_convert_list():
    assert convert_str([1, 2, 3]) == '[1, 2, 3]'
    
def test_convert_bad_float():
    assert convert_str('1.13Q') == '1.13Q'
    
def test_convert_unicode():
    assert convert_str('\u0030') == 0
    
def test_convert_negative():
    assert convert_str('-1') == -1
    
def test_covert_power():
    assert convert_str('\u00B2') == '\u00B2'


import pytest
from aiofiles.tempfile import NamedTemporaryFile

from aiocsv_utils.write import create_csv
from aiocsv_utils.write import write_csv_row, write_csv_rows


@pytest.mark.asyncio
async def test_create_csv():
    async with NamedTemporaryFile('w') as tmp:
        await create_csv(tmp.name, ['id', 'name', 'age', 'ssn'])
        with open(tmp.name, 'r') as f:
            results = f.read()
    assert results == 'id,name,age,ssn\n'
    
    
@pytest.mark.asyncio
async def test_write_csv_row():
    record = {'id': 1, 'name': 'John', 'age': 30, 'ssn': '111-22-3333'}
    headers = record.keys()
    async with NamedTemporaryFile('w') as tmp:
        await write_csv_row(tmp.name, record, headers)
        with open(tmp.name, 'r') as f:
            results = f.read()
    assert results == "1,John,30,111-22-3333\n"
    
    
@pytest.mark.asyncio
async def test_write_csv_rows():
    records = [{'id': 2, 'name': 'Jane', 'age': 25, 'ssn': '222-33-4444'},
               {'id': 3, 'name': 'Mike', 'age': 47, 'ssn': '000-11-2222'}]
    headers = records[0].keys()
    async with NamedTemporaryFile('w') as tmp:
        await write_csv_rows(tmp.name, records, headers)
        with open(tmp.name, 'r') as f:
            results = f.read()
    assert results == "2,Jane,25,222-33-4444\n3,Mike,47,000-11-2222\n"
    
    
@pytest.mark.asyncio
async def test_write_csv_row_to_headers():
    headers = ['id', 'name', 'age', 'ssn']
    record = {'age': 30, 'id': 1, 'name': 'John', 'ssn': '111-22-3333'}
    async with NamedTemporaryFile('w') as tmp:
        await create_csv(tmp.name, headers)
        await write_csv_row(tmp.name, record, headers)
        with open(tmp.name, 'r') as f:
            results = f.read()
    assert results == "id,name,age,ssn\n1,John,30,111-22-3333\n"
    

@pytest.mark.asyncio
async def test_write_csv_rows_to_headers():
    headers = ['id', 'name', 'age', 'ssn']
    records = [{'age': 25, 'id': 2, 'name': 'Jane', 'ssn': '222-33-4444'},
               {'id': 3, 'ssn': '000-11-2222', 'name': 'Mike', 'age': 47, }]
    async with NamedTemporaryFile('w') as tmp:
        await write_csv_rows(tmp.name, records, headers)
        with open(tmp.name, 'r') as f:
            results = f.read()
    assert results == "2,Jane,25,222-33-4444\n3,Mike,47,000-11-2222\n"


import typing as _typing


def convert_str(value: _typing.Any) -> float | int | bool | str:
    """Tries to coerce a value to float, int, bool and finally str.
    
    Parameters
    ----------
    value : Any
        Value to be coerced into float, int, or bool.
        
    Returns
    -------
    float | int | bool | str
        Returns coerced value if successful, or str value if it fails to coerce.
    """
    value = str(value)
    if value.count('.') == 1:
        try:
            return float(value)
        except ValueError:
            pass
    try:
        return int(value)
    except ValueError:
        pass
    if value == 'True':
        return True
    if value == 'False':
        return False
    return value